"""Defines the Singleton and UniqueInstance metaclasses.
"""
import hashlib
from typing import Any


class Singleton(type):
    """A standard singleton metaclass. Only one instance of a Singleton can be instantiated at
    runtime. If a request is made to instantiate a new Singleton, the existing Singleton is
    returned.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls.__name__ not in cls._instances.keys():
            cls._instances[cls.__name__] = super(Singleton, cls).__call__(
                *args, **kwargs
            )
        return cls._instances[cls.__name__]

    @classmethod
    def get_instances(mcs):
        return mcs._instances


class UniqueInstance(type):
    """A metaclass to provide instances with a unique identifier if no existing instance has
    that same identifier (otherwise return that instance).
    """

    _instances = {}

    def __call__(cls: Any, *args: Any, **kwargs: Any) -> Any:
        """Implements a factory pattern to return a new instance if there is no existing instance
        with the same ID; otherwise returns the existing instance with that ID. Skips __init__
        methods in all derived classes if returning an existing instance.
        """
        params = [i for i in list(args) + list(kwargs.values())]
        param_str = ".".join([str(i) for i in params])
        param_hex = hashlib.md5(param_str.encode()).hexdigest()

        cls._id = f"{cls.__name__}:{param_hex}"

        if cls._id not in cls._instances:
            self = cls.__new__(cls, *args, **kwargs)
            cls._instances[cls._id] = self
            cls.__init__(self, *args, **kwargs)

        return cls._instances[cls._id]

    def __init__(cls, *args, **kwargs):
        """Initialize the child class instance if returning a new instance."""
        super(UniqueInstance, cls).__init__(*args, **kwargs)

    @classmethod
    def get_instances(mcs):
        return mcs._instances



"""Define methods for mathematical operations.
"""
from __future__ import annotations

import math
import os
import random
import statistics
from typing import Any

from utilities import dictutils


def divide(a: int | float, b: int | float, r: int | float = None) -> float:
    """Return <a>/<b> if <b> is not zero otherwise return <r>.
    """
    try:
        return float(a) / float(b)
    except ZeroDivisionError:
        return 0.0 if r is None else r


def get_bradley_results(*, runs: int, r0: int, r1: int) -> dict:
    """Return a dictionary results for Bradley's runs tests. This method is called by methods
    that apply Bradley's runs tests for randomness in a list of elements.

        runs (int): total number of runs.
        r0 (int): number of runs above the median or of type A.
        r1 (int): number of runs below the median or of type B.

    See Bradley, (1968). Distribution-Free Statistical Tests, Chapter 12.
    """

    if r0 <= 10 and r1 <= 10:
        err_msg = (
            f"Parameters <r0> and <r1> must both be greater than or equal to 10. Received: "
            f"r0: {r0} and r1: {r1}."
        )
        raise ValueError(err_msg)

    expected_runs = 1 + int(divide((2 * r0 * r1), (r0 + r1), 0))
    numerator = 2 * r0 * r1 * (2 * r0 * r1 - r0 - r1)
    denominator = (r0 + r1) ** 2 * (r0 + r1 - 1)
    stddev = divide(numerator, denominator)
    zscore = divide(runs - expected_runs, stddev)

    return {
        "runs": runs,
        "runs-value-A": r0,
        "runs-value-B": r1,
        "expected-runs": expected_runs,
        "stddev": stddev,
        "z-score": zscore,
    }


def get_composites(limit: int, *, include_limit: bool = True) -> list:
    """Return a list of composite numbers up to the given limit. If optional <<include_limit>>
    is True (default), limit will be included in the returned numbers; if False, it will not be.
    """
    return get_primes_and_composites(limit=limit, include_limit=include_limit)[
        "composites"
    ]


def get_decimal_part(num: float) -> float:
    """Return the decimal part of the given num.
    """
    return abs(num) % 1


def get_percent_diff(*, accepted: int | float, measured: int | float) -> float:
    """Return the percent difference between the given values.
    """
    avg = 0.5 * (accepted + measured)
    return 100 * abs(accepted - measured) / avg


def get_percent_error(*, accepted: int | float, measured: int | float) -> float:
    """Return the percent error between the given values.
    """
    return 100 * abs((accepted - measured) / accepted)


def get_pvalue_from_zscore(zscore: float) -> float:
    """Given a z-score return the p-value.
    """
    return statistics.NormalDist().cdf(zscore) * 2 - 1


def get_primes(limit: int, *, include_limit: bool = True) -> list:
    """Return a list of prime numbers up to the given limit. If optional <<include_limit>> is
    True (default), limit will be included in the returned numbers; if False, it will not be.
    """
    return get_primes_and_composites(limit, include_limit=include_limit)["primes"]


def get_primes_and_composites(limit: int, *, include_limit: bool = True) -> dict:
    """The sieve of Eratosthanes. Return a dictionary with two keys, "composite" which maps to a
    list of composite numbers up to the given limit, and "prime" which maps to a list of prime
    numbers up the given limit. If optional <<include_limit>> is True (default), limit will be
    included in the returned numbers; if False, it will not be.
    """
    r = {"composites": set(), "primes": []}

    if (limit < 2) or (limit == 2 and not include_limit):
        return r
    elif limit == 2 and include_limit:
        r["primes"].append(2)
        return r
    elif limit == 3:
        r["primes"].append(2)
        if include_limit:
            r["primes"].append(3)
        return r

    for i in range(2, limit + 1):
        if i in r["composites"] or i in r["primes"]:
            continue

        for j in range(2 * i, limit + 1, i):
            if j in r["composites"] or j in r["primes"]:
                continue
            else:
                r["composites"].add(j)

        r["primes"].append(i)

    r["composites"] = list(r["composites"])

    return r


def get_random_capture(a: int | float, b: int | float) -> bool:
    """Return True if an internally generated random number in the range [0, 1] is between the given
    <a> and <b>.
    """
    return a < b and a <= get_random_number() < b


def get_random_mask(*, size: int, cut: float = 0.5, one_if_above: bool = True) -> list:
    """Return a flat 1-d mask of length <size>, containing only 1s and 0s.

    If optional <one_if_above> is:

        True.... each element has a 1 - <cut> chance of being ONE and a <cut> chance of being ZERO.
        False... each element has a 1 - <cut> chance of being ZERO and a <cut> chance of being ONE.
    """
    above_val = {1: True, 0: False}[one_if_above]
    below_val = {1: True, 0: False}[not one_if_above]

    return [
        above_val if i >= cut else below_val for i in get_random_number_list(size=size)
    ]


def get_random_number() -> float:
    """Return a random number in the range [0, 1].
    """
    return int.from_bytes(os.urandom(8), byteorder="big") / ((1 << 64) - 1)


def get_random_number_list(size: int) -> list[float]:
    """Return a list of random numbers each in the range [0, 1]. Makes a call to get_random_number,
    which see.
    """
    return [get_random_number() for _ in range(size)]


def get_zscore_from_pvalue(p: float) -> float:
    """Given a p-value return the z-score.
    """
    return statistics.NormalDist().inv_cdf((1 + p) / 2.0)


def is_normalized(table: dict, *, tol: float = 1e-5) -> bool:
    """Run a normalization tests against the give table.
    """
    return math.isclose(math.fsum(table.values()), 1, abs_tol=tol)


def is_random(
    r: list[int | float], *, zthresh: float = 1.960, sizethresh: int = 20
) -> dict:
    """Applies Bradly's run tests to the given list <r>, where len(set(r)) > 2, to determine if
    the values are randomized with significance level determined by the given z-score threshold.

    See:
        1. Bradley, (1968). Distribution-Free Statistical Tests, Chapter 12.
        2. https://www.itl.nist.gov/div898/handbook/eda/section3/eda35d.htm.
    """
    if len(set(r)) in [1, 2]:
        return is_random_binary_values(r=r, zthresh=zthresh, sizethresh=sizethresh)
    elif any([type(i) not in [float, int] for i in r]):
        err_msg = "All elements of the given list must be of type float or int."
        raise ValueError(err_msg)
    elif len(r) < sizethresh:
        err_msg = (
            f"The number of elements must exceed the size threshold: {sizethresh}. "
            f"Received: {len(r)} elements."
        )
        raise ValueError(err_msg)

    r_median = statistics.median(r)
    runs = r0 = r1 = 0  # r0 ascending, r1 descending

    for i in range(len(r)):
        if r[i] >= r_median and r[i - 1] < r_median:
            runs += 1
            r0 += 1
        elif r[i] < r_median and r[i - 1] >= r_median:
            runs += 1
            r1 += 1

    results = get_bradley_results(runs=runs, r0=r0, r1=r1)
    zscore = results["z-score"]

    results.update(
        {
            "len(r)": len(r),
            "z-score": zscore,
            "z-threshold": zthresh,
            "size-thresh": sizethresh,
            "num-distinct-values": len(set(r)),
            "runs": runs,
            "runs-value-A": r0,
            "runs-value-B": r1,
            "is-random": zscore > zthresh,
        }
    )

    return results


def is_random_binary_values(
    r: list[Any], *, zthresh: float = 1.960, sizethresh: int = 20
) -> dict:
    """Similar to is_random, which see. Intended for lists where each element is one of two
    possible values.

    See: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC6422539/
    """
    if len(r) < sizethresh:
        err_msg = (
            f"The number of elements must exceed the size threshold: {sizethresh}. "
            f"Received: {len(r)} elements."
        )
        raise ValueError(err_msg)
    elif len(set(r)) > 2:
        err_msg = (
            f"The given list must contain not more than two distinct values. Received: "
            f"{len(set(r))} distinct values."
        )
        raise ValueError(err_msg)

    runs = r0 = 1

    for i in range(1, len(r)):
        if r[i] != r[i - 1]:  # we are starting a new run
            runs += 1
            if r[i] == r[0]:
                r0 += 1

    r1 = runs - r0
    results = get_bradley_results(runs=runs, r0=r0, r1=r1)
    zscore = results["z-score"]

    results.update(
        {
            "len(r)": len(r),
            "z-score": zscore,
            "z-threshold": zthresh,
            "size-thresh": sizethresh,
            "num-distinct-values": len(set(r)),
            "runs": runs,
            "runs-value-A": r0,
            "runs-value-B": r1,
            "is-random": zscore > zthresh,
        }
    )

    return results


def get_integers_with_random_step(*, start: int, stop: int, maxstep: int) -> list[int]:
    """Return a list of non-repeating random integers whose smallest-possible element is
    <start> and whose largest-possible element is <stop>, with the step between each element (
    when are they written in order from largest to smallest) being an integer randomly-selected
    from the range [1, maxstep].

    NOTES:
        1. The order of the elements in the returned list is not guaranteed.
        2. This method will raise ValueError if <<start>> is not less than <<stop>> or if
            <<maxstep>> < 2.
    """
    if start >= stop:
        err_msg = (
            f"Parameter <start> must be less than parameter <stop>. Received: start: "
            f"{start}; stop: {stop}."
        )
        raise ValueError(err_msg)

    if maxstep < 2:
        err_msg = f"Parameter <maxstep> must be greater than or equal to 2. Received: {maxstep}."
        raise ValueError(err_msg)

    n = start - 1  # makes it possible for <start> to be included in the list
    r = []  # result list

    while n < stop:
        step = random.randint(1, maxstep)
        next_n = n + step
        if next_n < stop:
            r.append(next_n)
            n = next_n
        else:
            r.append(stop)
            n = stop
    return r


def get_normalized_table(table: dict[Any:float]) -> dict[Any:float]:
    """Return a normalized table based on the given table, which is not modified.
    """
    z = math.fsum(table.values())
    return {k: v / z for k, v in table.items()}


def normalize_table(table: dict[Any:float]) -> None:
    """Normalize the given table in place.
    """
    t_new = get_normalized_table(table)
    for k in t_new:
        table[k] = t_new[k]


def select_from_table(table: dict) -> Any:
    """Return a randomly-selected key from the given table, for which it is assumed the keys are
    ordered such that the mapped values are in order from smallest to largest, with the smallest
    value being 0.0 and the largest being 1.0 (exact). Calls select_from_table_by_cut, which see.
    """
    return select_from_table_by_cut(table, cut=get_random_number())


def select_from_table_by_cut(table: dict, *, cut: int | float) -> Any:
    """Return a key from the given table, for which it is assumed the keys are
    ordered such that the mapped values are in order from smallest to largest, with the smallest
    value being greater than 0.0 and the largest being 1.0 (exact). Under this assumption,
    the first key for which <cut> is NOT greater than the mapped value is returned. If no such
    key is found, return None.
    """
    if cut > 1:
        err_msg = f"Parameter cut cannot be greater than 1. Received: {cut}."
        raise ValueError(err_msg)

    dictutils.reorder_by_value(table)

    for k, v in table.items():
        if cut <= v:
            return k


def sign(a: Any) -> float:
    """Return the sign (+1 or -1) of the given input.
    """
    return (lambda i: math.copysign(1, i))(a)


def xor(a: Any, b: Any) -> bool:
    """Return a xor b.
    """
    return bool(a) != bool(b)



"""Define utility methods to facilitate user interactions.
"""
import subprocess
import sys
import termios
import tty


def any_key_continue() -> None:
    while True:
        option = get_char()
        if option.lower() == "q":
            confirm_quit()
        else:
            return


def clear_screen() -> None:
    subprocess.run(["clear"])


def confirm_quit() -> None:
    msg = "Are you sure you want to exit? [Y/y] to confirm."
    display_message(msg)
    confirm = get_char()
    if confirm.lower == "y":
        sys.exit()
    else:
        any_key_continue()


def display_message_and_wait(msg: str) -> None:
    clear_screen()
    display_message(msg=msg)
    any_key_continue()


def display_message(msg: str) -> None:
    print(f"\n{'*'*12}\n{msg}\n{'*'*12}\n")


def get_char() -> str:
    """Get character input from the keyboard."""

    def _get_char():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch

    return _get_char()


def get_option_as_int() -> int:
    """Get integer option input from keyboard."""
    msg = "<<< Enter option or [Q/q] to quit: >>>"
    display_message(msg)
    option = get_char()
    if option.lower() == "q":
        confirm_quit()
    try:
        option = int(option)
    except ValueError:
        msg = f"Invalid option: {option}."
        display_message(msg=msg)
    else:
        return option


def show_progress_bar(
    iteration: int,
    total: int,
    prefix: str = "",
    suffix: str = "",
    decimals: int = 1,
    length: int = 100,
    fill: str = "█",
    end_char: str = "\r",
) -> None:
    """Create a progress bar at the command line.

    https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    bar_length = int(length * iteration // total)
    bar = fill * bar_length + "-" * (length - bar_length)
    print(f"\r{prefix} |{bar}| {percent}% {suffix}", end=end_char)

    if iteration == total:
        print()



"""This module defines decorators.
"""
import logging
import time

from utilities import dictutils

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def timing(screen=False, verbose=False):
    """For a decorated method, write two log statements recording (a) when the method was called
    and optionally its input arguments (if verbose==True), and (b) when the method finished and
    the total duration in seconds. If optional <<screen>> is True, output is directed to stdout
    instead of the log.

    Usage:

        @timing()
        def decorated_method():
            [do things...]

        @timing(verbose=True)
        def decorated_method(a, b, c=None):
            [do things...]

    """

    def decorator(func):
        def wrapper(*args, **kwargs):
            name = f"Function {func.__name__}"
            started_at = f"started at {time.strftime('%H:%M:%S')}"

            if verbose:
                started_at += f"\n\t{name} input arguments:\n\t\t\t{dictutils.get_string(locals())}"

            msg = f"{name} {started_at}"
            if screen:
                print(msg)
            else:
                logger.info(msg)

            clock_start = time.time()
            result = func(*args, **kwargs)
            clock_end = time.time()

            finished_at = f"finished at {time.strftime('%H:%M:%S')}"
            with_duration = f"with duration {(clock_end - clock_start):.3f} seconds"

            msg = f"{name} {finished_at} {with_duration}."

            if screen:
                print(msg)
            else:
                logger.info(msg)

            return result

        return wrapper

    return decorator



"""Define general utility methods not easily classified for any of the other
utility modules.
"""

from typing import Any


def convert_to_type(input_value: Any, target_type: Any) -> Any:
    """Convert the given object to the given type. Raises TypeError if the conversion fails.

    Note: For converting type defaultdict to type dict, please see dictutils.get_pure_dict.
    """
    if isinstance(input_value, target_type):
        return input_value

    # Handle special cases, if necessary.

    if isinstance(input_value, str) and isinstance(target_type, list):
        return [input_value]

    try:
        return target_type(input_value)
    except TypeError as e:
        err_msg = f"Failed to convert {input_value} to {target_type}."
        raise TypeError(err_msg) from e



"""Define utility methods for working with dates and datetime objects.
"""
from __future__ import annotations

import calendar
from datetime import datetime, timedelta

DAYS_PER_CENTURY = 36_525
JAN_1_2000_NOON = 2_451_545  # Julian day number.
SHORT_DATE_FORMAT = "%Y-%m-%d"  # ISO-8601.


def get_clock_time() -> str:
    return datetime.now().strftime("%H:%M:%S")


def check_date_format(*, date: str, dtformat: str = None) -> bool:
    """Return True if the given <<date>> is formatted according to
    <<dt_format>>, otherwise False.
    """
    dtformat = SHORT_DATE_FORMAT if dtformat is None else dtformat
    try:
        datetime.strptime(date, dtformat)
    except ValueError:
        return False
    return True


def get_date_ahead(r: datetime = datetime.now(), *, days_ahead: int) -> datetime:
    return r + timedelta(days_ahead)


def get_date_short_str(*, date: datetime) -> str:
    return date.strftime(SHORT_DATE_FORMAT)


def get_datestamp(date: datetime = None, with_time: bool = False) -> str:
    now = datetime.now() if date is None else date
    if with_time:
        datestamp = now.isoformat().split(".")[0].replace(":", "-")
    else:
        datestamp = now.isoformat().split("T")[0]
    return datestamp


def get_datetime_instance(*, date: str) -> datetime:
    """Return a datetime instance."""
    return datetime.strptime(date, SHORT_DATE_FORMAT)


def get_julian_centuries(*, date: datetime) -> float:
    """Convert the given date into the number of Julian centuries Jan 1, 2000, Noon.
    """
    julian_date = get_julian_day(date=date)
    julian_centuries = (julian_date - JAN_1_2000_NOON) / DAYS_PER_CENTURY

    return julian_centuries


def get_julian_day(*, date: datetime) -> float:
    """Convert the given date into the Julian day number.

    Note: year is restricted as follows:

        1901 <= year <= 2099
    """
    year = date.year

    if not (1901 <= year <= 2099):
        raise ValueError(
            f"The year must between 1901 and 2099, inclusive. Received: {year}."
        )

    month = date.month
    day = date.day
    hour = date.hour
    minute = date.minute
    second = date.second
    universal_time = hour + minute / 60 + second / 3600
    ut_fraction = universal_time / 24

    part_a = 367 * year
    part_b_numerator = 7 * (year + int((month + 9) / 12))
    part_b = int(part_b_numerator / 4)
    part_c = int((275 * month) / 9)

    julian_day = part_a - part_b + part_c + day + 1_721_013.5 + ut_fraction

    return julian_day


def get_last_date_of(r: datetime, *, step: str) -> datetime:
    """If step is one week, return the date for the first Sunday on or after the given date <r>.
    If step is one month, return the date for the last day of the month containing <r>.
    """
    if step.strip().lower() == "week":
        days_left = 7 - r.isoweekday()
        last_date = r + timedelta(days_left)
    elif step.strip().lower() == "month":
        last_day = calendar.monthrange(r.year, r.month)[-1]
        last_date = datetime(r.year, r.month, last_day)

    return last_date



"""Define utility methods for working with the git (gitpython) module.

    $> pip install gitpython
"""
from __future__ import annotations

from pathlib import Path

import git


def is_git_repo(path: Path) -> bool:
    """Is the given path a git repo? If yes, return True. If no, return False."""
    try:
        _ = git.Repo(path).git_dir  # noqa
        return True
    except git.exc.InvalidGitRepositoryError:  # noqa
        return False



"""Define utilities that work with matplotlib.pyplot.
"""
from __future__ import annotations

import time
from pathlib import Path

import matplotlib
import matplotlib.pyplot as pyplot
import numpy

from utilities.genutils import convert_to_type

matplotlib.use("TkAgg")


def plotdata(
    *,
    data: numpy.ndarray,
    dependent_params: dict,
    independent_params: dict,
    title: str,
    show_plot_time: int | None = None,
    filepath: Path | None = None,
) -> None:
    """Plot the given data based on the plot indexes, y_limits and x_limits.

    Notes:
        (1) dependent_params must contain keyword "columns" mapping to a list of integers.

        (2) independent_params must contain keyword "column" mapping to an integer. The specified
        column will be plotted as the independent variable.

        (3) The greatest integer in dependent_params "columns" list must not be greater than the
        number of columns in the data. The columns specified will be plotted as dependent
        variables, all overlaid on the same plot.

        (4) ylim and xlim are each a list of two real numbers, specifying the upper
        and lower limits displayed on the plot for the x and y axes.

        (5) Optional show_plot_time provides the integer number of seconds to display the plot
        in a separate window before closing the window and returning control to the
        interpreter. If this parameter is not provided, the plot window is not generated.

        (6) Optional filepath provides string or Path object specifying the path to which an image
        of the plot should be saved. If this parameter is not provided, no file is saved.

        (7) If both show_plot_time and filepath are not provided, nothing is done.
    """
    if data is None or (show_plot_time is None and filepath is None):
        return

    pyplot.figure(num=1, dpi=150, figsize=(8, 5))
    pyplot.rcParams["font.family"] = "serif"
    pyplot.rcParams["figure.facecolor"] = "white"
    pyplot.rcParams["savefig.facecolor"] = "white"

    dep_cols = dependent_params["columns"]
    dep_axis_label = dependent_params["axis_label"]
    dep_labels = dependent_params["curve_labels"]
    dep_lims = dependent_params["lims"]

    ind_col = independent_params["column"]
    ind_axis_label = independent_params["axis_label"]
    ind_lims = independent_params["lims"]

    dep_cols = convert_to_type(dep_cols, list)
    dep_labels = convert_to_type(dep_labels, list)

    ind_vals = [d[ind_col] for d in data]

    for dep_col, dep_label in zip(dep_cols, dep_labels):
        dep_vals = [d[dep_col] for d in data]
        pyplot.plot(ind_vals, dep_vals, label=dep_label, markersize=1.25)

    pyplot.title(title)
    pyplot.legend(loc="lower right")
    pyplot.grid(True, color="lightgray")
    pyplot.ylabel(dep_axis_label)
    pyplot.xlabel(ind_axis_label)
    pyplot.ylim(dep_lims)
    pyplot.xlim(ind_lims)

    if show_plot_time is not None:
        pyplot.show(block=False)
        time.sleep(show_plot_time)

    if filepath is not None:
        pyplot.savefig(filepath, transparent=False, bbox_inches="tight")



"""Define utility methods to read from and write to various filetypes.
"""
from __future__ import annotations

import copy
import csv
import json
import logging
import pickle
import shutil
from collections import defaultdict
from datetime import datetime
from itertools import zip_longest
from pathlib import Path
from typing import Any
from xml.etree import ElementTree as eTree

import blosc
import yaml

from utilities import dictutils, listutils, pathutils

logger = logging.getLogger(__name__)

READFILE_EXTENSIONS = [".csv", ".dat", ".json", ".yaml", ".yml"]


def combine_json_files(filepaths: list[Path], *, saveto: Path) -> None:
    """Takes in a list of filepaths, assumed to be all type json, and combines them into a
    single json file at the specified filepath.
    """
    suffixes = list(set([f.suffix.lower() for f in filepaths]))

    if not all([confirm_filetype(f, "json") for f in filepaths]):
        raise ValueError(
            f"All filepaths must be of type JSON. Received the following file "
            f"extensions: {suffixes}."
        )

    if not confirm_filetype(saveto, "json"):
        raise ValueError(
            f"Parameter <<savetojson>> must have extension 'json'. Received:"
            f" {saveto}."
        )

    savetoparent = filepaths[0].parent

    if not savetoparent.is_dir():
        logger.info(f"Directory {savetoparent} does not exist. Creating it.")
        savetoparent.mkdir(parents=True, exist_ok=True)

    data = {}
    for f in filepaths:
        if f.is_file():
            filedata = readfile(f)
            if filedata:
                data.update(filedata)
            else:
                logger.warning(f"File {f} has no data. Nothing to do.")
        else:
            logger.warning(f"File {f} does not exist. Skipping it.")
            continue

    writefile(saveto, data=data)


def confirm_filetype(filepath: Path, extension: str) -> bool:
    """Return boolean True if the given filepath has the specified extension.
    """
    if not extension.startswith("."):
        extension = f".{extension}"
    suffix = filepath.suffix
    return suffix.lower() == extension.lower()


def create_backup_file(filepath: Path) -> None:
    """Create a backup of the given file, appending a datestamp to the filename."""
    if not filepath.exists():
        logger.warning(f"{filepath} does not exist. Nothing to do.")
        return

    today = datetime.today().strftime("%Y-%m-%d-%H-%M-%S")
    fileparent = filepath.parent
    filestem = filepath.stem
    filesuffix = filepath.suffix
    newfilename = f"{today}-{filestem}{filesuffix}"
    newfilepath = f"{fileparent}/{newfilename}"

    try:
        shutil.copy2(filepath, newfilepath)
    except FileNotFoundError:
        logger.warning(
            f"Error copying source {filepath} to destination {newfilepath}: {filepath} "
            f"does not exist. Nothing to do."
        )
        return

    logger.info(f"{filepath} copied to {newfilepath}.")


def __handle_csv(
    filepath: Path,
    *,
    as_rows: bool,
    data: dict | defaultdict | list,
    exclude_keys: list,
    header_row: list[Any] = None,
    include_keys: bool,
    max_rows: int,
    mode: str,
    use_legacy: bool = False,
) -> None:
    """Call the appropriate method to write data to a CSV file. If use_legacy is True,
    only data, max_rows, and mode are relevant-- all other keys are ignored.
    """
    if use_legacy:
        write_csvfile_from_dicts_legacy(
            filepath, data=data, max_rows=max_rows, mode=mode
        )

    if type(data) in [dict, defaultdict]:
        if as_rows:
            write_csvfile_d2r(
                filepath,
                data=data,
                exclude_keys=exclude_keys,
                header_row=header_row,
                include_keys=include_keys,
                mode=mode,
            )
        else:
            write_csvfile_d2c(filepath, data=data, exclude_keys=exclude_keys, mode=mode)
    elif type(data) is list:
        write_csvfile_l2rc(
            filepath, as_rows=as_rows, data=data, header_row=header_row, mode=mode
        )


def read_csvfile(filepath: Path, delimiter: str = ",") -> list[list]:
    """Takes in a filepath and loads the data in the CSV file into a list of lists, with one outer
    list for each row of the file. Each inner list contains the individual data elements for a
    particular row.
    """
    with open(filepath, "r", newline="\n") as f:
        lines = f.readlines()

    return [line.strip().split(delimiter) for line in lines]


def read_datfile(filepath: Path) -> dict:
    """Takes in filepath and loads the data in the DAT file into a dictionary. It is assumed the
    DAT file is blosc-compressed pickle of a Python dictionary object.
    """
    with open(filepath, "rb") as f:
        try:
            data = pickle.loads(blosc.decompress(f.read()))
        except EOFError as e:  # no other blosc exception?
            logger.warning(
                f"Could not load cached data from {filepath}: {e.__str__()}. "
                f"Returning empty dictionary."
            )
            data = {}

    return data


def readfile(filepath: Path) -> list | dict | None:
    """Takes in a Path instance, calls the appropriate read method based on the filename
    extension, and returns the content of the file.
    """
    if not filepath.is_file():
        logger.warning(f"{filepath} does not exist. Nothing to do.")
        return

    if filepath.suffix in READFILE_EXTENSIONS:
        return {
            ".csv": read_csvfile,
            ".dat": read_datfile,
            ".json": read_jsonfile,
            ".yaml": read_yamlfile,
            ".yml": read_yamlfile,
        }[filepath.suffix](filepath)
    else:
        raise ValueError(
            f"Parameter <<filepath>> must have one of the following extensions: "
            f"{listutils.get_string(READFILE_EXTENSIONS)}. Received: {filepath}."
        )


def read_jsonfile(filepath: Path) -> list | dict:
    """Takes in a Path instance and loads the data in the file into either a list or dictionary.
    """
    with open(filepath, "r") as f:
        try:
            data = json.load(f)
        except json.decoder.JSONDecodeError as e:
            err_msg = f"Could not read file {filepath}."
            raise json.decoder.JSONDecodeError(err_msg, str(filepath), 0) from e
    return data


def read_yamlfile(filepath: Path) -> list | dict:
    """Takes in a Path instance and loads the data in the file into either a list or dictionary.
    """
    with open(filepath, "r") as f:
        data = yaml.safe_load(f)

    return data


def set_mode(filepath: Path) -> str:
    """Set the mode flag based on the incoming filepath. If the file exists, mode is set to
    <<a>>. If it does not exist, mode is set to <<w>>.
    """
    return {True: "a", False: "w"}[filepath.is_file()]


def write_csvfile_from_dicts_legacy(
    filepath: Path,
    *,
    data: dict,
    exclude_keys: list = None,
    max_rows: int = None,
    mode: str = "w",
) -> None:
    """
    WARNING: This is a legacy method that will be removed by 2025-01-01.

    Write the given data dictionary to one or more CSV files. Each k-v pair of the data
    dictionary represents one row of data, and v is itself a dictionary for which the keys
    correspond to the column headers. It is assumed that all such dictionaries have the same
    key-set.

    NOTES:
        1. The values in data must all be of the same type, either dict or list. If all are of type
            list, this method calls write_csvfile_lists. If values are not all dict or list,
            nothing will be written to disk.

        2. If max rows is given, write this many rows (including the header) in a csv file,
            and create as many csv files as necessary to write the entire data dictionary. A
            positive-integer suffix (starting with 1) is added to each filename based on the given
            filepath.

        3. Any key names provided in optional parameter <<exclude_keys>> will be ignored.
    """
    if not data:
        return

    data_value_types = list(set([type(v) for v in data.values()]))

    if all([dvt is list for dvt in data_value_types]):
        return write_csvfile_d2r(
            filepath, data=data, mode=mode, exclude_keys=exclude_keys
        )
    elif not all([dvt is dict for dvt in data_value_types]):
        logger.warning(
            f"All values in parameter <<data>> must be of type dict. Received types "
            f"{data_value_types}. Data WAS NOT WRITTEN to disk."
        )
        return

    def _write_csvdata(
        data: dict, filepath: Path, mode: str, exclude_keys: list[str] = None
    ):
        exclude_keys = [] if exclude_keys is None else exclude_keys

        with open(filepath, mode, newline="") as f:
            # Get the column names from the first row of the data map--- we assume all rows have the
            # same keys--- but only if mode == "w". If writing in append mode, assume that we
            # already have the header row.

            column_names = [k for k in list(data.values())[0] if k not in exclude_keys]
            writer = csv.DictWriter(f, column_names)

            if mode == "w":
                writer.writeheader()

            for row_key, row_data in data.items():
                row = {}
                for col_name in column_names:
                    try:
                        row[col_name] = row_data[col_name]
                    except KeyError:
                        logger.warning(
                            f"Column name missing from row data: {col_name}."
                        )
                        row[col_name] = "missing"
                writer.writerow(row)

    if max_rows is not None:
        if not isinstance(max_rows, int) or max_rows < 2:
            raise ValueError(
                f"Parameter <<max_rows>> must be a positive integer >= 2. "
                f"Received {max_rows}."
            )

    data = dictutils.get_pure_dict(data)
    exclude_keys = [] if exclude_keys is None else exclude_keys
    working_filepath = Path(copy.deepcopy(filepath))

    if max_rows:
        filecounter = 1
        for subdata in dictutils.get_smaller_dicts(data, max_keys=max_rows):
            _write_csvdata(subdata, working_filepath, mode)
            filecounter += 1
            working_filepath = pathutils.add_suffix_to_path(
                path=filepath, suffix=str(filecounter)
            )
    else:
        _write_csvdata(data, working_filepath, mode, exclude_keys)


def write_csvfile_d2c(
    filepath: Path,
    *,
    data: dict[str : list[float]] | defaultdict[str : list[float]],
    exclude_keys: list[Any] = None,
    mode: str = "w",
) -> None:
    """Write the given data to a CSV file. Assume the given data represents columns in the CSV
    file, with each key being the column header and each value being the data in that column. If
    optional exclude_keys is provided, those keys will be ignored when parsing data.

    :param filepath: where the CSV file will be written
    :param data: the data to write to the CSV file
    :param mode: python write mode; default is "w"
    :param exclude_keys: optional; keys in this list will be ignored
    :return:
    """
    data = dictutils.get_pure_dict(data)
    if exclude_keys is None:
        exclude_keys = []
    column_names = [k for k in list(data.keys()) if k not in exclude_keys]
    column_lists = [v for k, v in data.items() if k not in exclude_keys]
    rows = zip_longest(*column_lists, fillvalue="")

    with open(filepath, mode, newline="") as f:
        writer = csv.writer(f)
        writer.writerow(column_names)
        for row in rows:
            writer.writerow(row)


def write_csvfile_d2r(
    filepath: Path,
    *,
    data: dict[str : list[float]] | defaultdict[str : list[float]],
    exclude_keys: list[str] = None,
    header_row: list[Any] = None,
    include_keys: bool = False,
    mode: str = "w",
) -> None:
    """Write the given data to a CSV file. Assume that each kv-pair in data represents one row
    of data. If include_keys is True, the key in each kv-pair will be included as the first
    element on the row. The value in each pair will comprise the (remaining) elements on the row.
    If optional exclude_keys is given, those keys in data will be ignored.

    :param filepath: where the CSV file will be written
    :param data: the data to write to the CSV file
    :param exclude_keys: optional; ignore these keys in data
    :param header_row: optional; header row
    :param include_keys: optional; if True, include the key as the first element on a row
    :param mode: python write mode; default is "w"
    :return:
    """
    data = dictutils.get_pure_dict(data)

    if exclude_keys is None:
        exclude_keys = []
    data_rows = []

    for key, values in data.items():
        if key in exclude_keys:
            continue
        this_row = []
        if include_keys:
            this_row.append(key)
        this_row.extend(values)
        data_rows.append(this_row)

    with open(filepath, mode, newline="") as f:
        writer = csv.writer(f)
        if header_row:
            writer.writerow(header_row)
        for row in data_rows:
            writer.writerow(row)


def write_csvfile_l2rc(
    filepath: Path,
    *,
    as_rows: bool = True,
    data: list[list[float]],
    header_row: list[Any] = None,
    mode: str = "w",
) -> None:
    """Write the given data to a CSV file. If as_rows is True, assume each list in data represents
    one row, and if False assume each list in data represents one column. If provided, include the
    optional header_row as the first line of the CSV file.

    :param filepath: where the CSV file will be written
    :param as_rows: if True assume data lists correspond to rows, otherwise columns
    :param data: the data to write to the CSV file
    :param mode: python write mode; default is "w"
    :param header_row: optional header row
    :return:
    """
    if as_rows:
        rows = data
    else:
        rows = zip_longest(*data, fillvalue="")

    with open(filepath, mode, newline="") as f:
        writer = csv.writer(f)
        if header_row:
            writer.writerow(header_row)
        for row in rows:
            writer.writerow(row)


def write_datfile(filepath: Path, *, data: dict) -> None:
    """Takes in a filepath and a data dictionary, blosc-compresses a pickle of the data,
    and writes the compressed pickle to disk as a DAT file.
    """
    if filepath.is_file():
        current = readfile(filepath)
        data.update(current)

    with open(filepath, "wb") as f:
        f.write(blosc.compress(pickle.dumps(data)))


def writefile(
    filepath: Path,
    *,
    as_rows: bool = True,
    data: dict | defaultdict | list,
    exclude_keys: list[Any] = None,
    header_row: list[Any] = None,
    include_keys: bool = False,
    max_rows: int = None,
    mode: str = "w",
    root: eTree.Element = None,
) -> None:
    """Takes in a stringutils or Path instance filepath, a data dictionary or list, and one or
    more optional parameters, and calls the appropriate write method based on the filepath
    extension.

    NOTES:

        (1) If the filepath extension is ".csv":

            (a) If parameter <<as_lists>> is False, parameter <<exclude_keys>> is ignored.
            (b) If parameter <<as_rows>> is False, parameter <<col_indexes>> is ignored.

        (2) If the filepath extension is NOT ".csv", parameters <<as_lists>>, <<as_rows>>,
            <<exclude_keys>>, and <<col_indexes>> are ignored.
        (3) If the filepath extension is ".dat", parameter <<data>> must be of type <<dict>>.
        (4) If the filepath extension is ".dat" OR ".xml", parameter <<mode>> is ignored.
        (5) If the filepath extension is ".xml", parameter <<root>> is required.
        (6) If the filepath extension is NOT ".xml", parameter <<root>> is ignored.
    """
    suffix = filepath.suffix

    if suffix == ".csv":
        __handle_csv(
            filepath,
            as_rows=as_rows,
            data=data,
            exclude_keys=exclude_keys,
            header_row=header_row,
            include_keys=include_keys,
            max_rows=max_rows,
            mode=mode,
        )
    elif filepath.suffix == ".dat":
        write_datfile(filepath, data=data)
    elif filepath.suffix == ".json":
        write_jsonfile(filepath, data=data, mode=mode)
    elif filepath.suffix in [".yaml", "yml"]:
        write_yamlfile(filepath, data=data, mode=mode)
    elif filepath.suffix == ".xml":
        write_xmlfile(filepath, root=root)
    else:
        logger.warning(
            f"Parameter <<max_rows>> must be a positive integer >= 2. Received "
            f"{max_rows}. Data WAS NOT WRITTEN to disk."
        )
        return

    logger.info(f"Data written to file: {filepath}.")


def write_jsonfile(filepath: Path, *, data: dict | list, mode: str = "w") -> None:
    """Takes in Path instance, a data dictionary or list, optional mode
    dictionary or listutils to the specified pathutils in JSON format. If appending, the incoming
    data must be of the same type as that extracted from the given filepath. If it is not,
    this method raises ValueError.
    """
    if mode == "a":
        try:
            existing_data = read_jsonfile(filepath)
        except FileNotFoundError:
            logger.warning(f"Filepath {filepath} not found.")
            pass
        else:
            if type(data) != type(existing_data):
                logger.warning(
                    f"Cannot merge data objects of different types. "
                    f"Received {type(data)} with append mode, but existing data is "
                    f"of type {type(existing_data)}. Data WAS NOT WRITTEN to disk."
                )
                return

            if isinstance(existing_data, list):
                data += existing_data
            elif isinstance(existing_data, dict):
                data = {**existing_data, **data}

    with open(filepath, "w") as f:
        json.dump(data, f)


def write_xmlfile(filepath: Path, *, root: eTree.Element) -> None:
    """Takes in a stringutils or Path instance filepath and an eTree.Element object, which should
    be the root of an eTree.ElementTree object, and coerces it and its children into
    nicely-formatted XML and writes the data to file.

    NOTE: Relies on an internally-defined recursive method.

    See: https://stackoverflow.com/a/65808327
    """

    def _format_xml(
        element: eTree.Element,
        parent: eTree.Element = None,
        index: int = -1,
        depth: int = 0,
    ) -> None:
        """Format the incoming element and its children into nicely-formatted XML. Recursive.
        """
        for i, node in enumerate(element):
            _format_xml(node, element, i, depth + 1)  # Recursive call.

        if parent is not None:
            if index == 0:
                parent.text = "\n" + ("\t" * depth)
            else:
                parent[index - 1].tail = "\n" + ("\t" * depth)

            if index == len(parent) - 1:
                element.tail = "\n" + ("\t" * (depth - 1))

    _format_xml(root)
    tree = eTree.ElementTree(root)
    tree.write(filepath)


def write_yamlfile(filepath: Path, *, data: dict | list, mode: str = "w") -> None:
    """Takes in a data dictionary or listutils, a stringutils or Path instance, and an optional
    boolean flag and writes the data dictionary or list to the specified path in YAML
    format. If appending, the incoming data must be of the same type as that extracted from the
    given filepath. If it is not, this method raises ValueError.
    """
    if mode == "a":
        try:
            existing_data = read_yamlfile(filepath)
        except FileNotFoundError:
            logger.warning(
                f"Append mode is {mode}, but {filepath} was not found. Data WAS NOT "
                f"WRITTEN to disk."
            )
            return
        else:
            if type(data) != type(existing_data):
                logger.warning(
                    f"Cannot merge data objects of different types. Received"
                    f" {type(data)} "
                    f"with append mode, but existing data is of type {type(existing_data)}. Data "
                    f"WAS NOT WRITTEN to disk."
                )
                return

            if isinstance(existing_data, list):
                data += existing_data
            elif isinstance(existing_data, dict):
                data = {**existing_data, **data}

    with open(filepath, "w") as f:
        yaml.dump(data, f)



"""Define utility methods for working with Path objects.
"""
from __future__ import annotations

import os
import shutil
import sys
from pathlib import Path

from utilities.genutils import convert_to_type


def add_extension(filepath: Path, ext: str) -> Path:
    """Add an extension to the Path object if it does not already have one."""
    if not filepath.suffix:
        return Path(str(filepath) + ext)
    else:
        return filepath


def copy_all_files(*, sourcepath: Path, targetpath: Path):
    """Copy all files in the source directory to the target directory."""

    def _get_pairs(target_path):
        pairs = [
            (item, target_path / item.name)
            for item in sourcepath.glob("*")
            if item.is_file()
        ]

        pyversion = int(sys.version_info[0]) + int(sys.version_info[1]) / 10

        if pyversion <= 3.7:
            pairs = [(str(p[0]), str(p[1])) for p in pairs]
        elif pyversion < 3.0:
            err_msg = "Python version must be 3+."
            raise ValueError(err_msg)

        return pairs

    for pair in _get_pairs(targetpath):
        shutil.copy(*pair)


def count_files(path: Path):
    """Return the integer number of files in the given directory and all subdirectories."""
    return sum(1 for _, _, files in os.walk(path) for _ in files)


def confirm_filetype(filepath: str | Path, expected_extensions: str | list) -> bool:
    """Return boolean True if the given filepath has an extension that matches any of the given
    expected extensions."""
    filepath = convert_to_type(filepath, Path)
    expected_extensions = convert_to_type(expected_extensions, list)
    suffix = filepath.suffix
    return any(suffix.lower() == expect.lower() for expect in expected_extensions)


def convert_to_path(path: str | Path) -> Path:
    """Convert the given path to type Path (in place) if it is type str. Otherwise, do nothing."""
    return Path(path) if path is isinstance(path, str) else path


def set_full_filepath(
    *,
    target_path: str | None,
    root_path: str,
    default_path: str,
    filename: str,
    parents: bool = True,
    exist_ok: bool = True,
):
    """Set the full filepath for the given filename. If path is None, use
    default_path.
    """
    if target_path is None:
        filepath = Path(root_path, default_path, filename)
    else:
        Path(root_path, target_path).mkdir(parents=parents, exist_ok=exist_ok)
        filepath = Path(root_path, target_path, filename)

    return filepath


def add_suffix_to_path(*, path: Path, suffix: str) -> Path:
    """Add a suffix to a Path object.

    Example:

        Given pathutils: "../../output/data.csv", as a Path object.
        Given suffix: "-01", as a stringutils.

        Return: "../../output/data-01.csv", as a Path object.
    """
    parent = str(path.parent)
    pathname, extension = str(path.name).split(".")
    pathname = "\\" + f"{pathname}-{suffix}.{extension}"

    return Path(parent + pathname)



"""Define utility methods for working with dictionaries.
"""
from __future__ import annotations

import heapq
import logging
from collections import defaultdict
from typing import Any

from utilities import listutils

logger = logging.getLogger(__name__)


def confirm_keys(d: dict, *, keys: list) -> None:
    """Return boolean True if each of the elements in list keys is a key in the given dictionary."""
    if all([k in list(d.keys()) for k in keys]):
        return
    else:
        err_msg = (
            f"Parameter <d> does not contain all required keys.\n"
            f"\tRequired keys: {listutils.get_string(keys)}"
            f"\tReceived keys: {listutils.get_string(list(d.keys()))}."
        )
        raise ValueError(err_msg)


def compare_dicts(d1: dict, d2: dict) -> dict:
    """
    Return a dictionary of dictionaries comparing the given d1 and d2 dictionaries on the
    basis of the k-v pairs of each but without regard to the ordering of the keys in either.

    The returned dictionary contains four outer keys (as strings) that map to the result
    dictionaries:

        * "unique-to-d1": maps to a dictionary of k-v pairs where the keys are unique to d1.

        * "unique-to-d2": maps to a dictionary of k-v pairs where the keys are unique to d2.

        * "conflicting": maps to a dictionary of k-v pairs where the keys are common to d1 and d2
        but for each key the value in d1 does not equal the value in d2. Each key is mapped to a
        tuple that holds these two unequal values, with the first element coming from d1 and the
        second element coming from d2.

        * "identical": maps to a dictionary of k-v pairs where the keys are common to d1 and d2
        and for each key the value in d1 equals the value in d2, and is the value mapped by that
        same key in the returned dictionary.

    Notes:
        1. If the returned dictionary is constructed correctly, at least two of its
        result dictionaries will be populated, and each key of the given d1 and d2 dictionaries
        will appear exactly once in exactly one of the result dictionaries.

        2. If the returned dictionary is constructed correctly, there cannot be any tuple in
        the "conflicting" data for which the first and second elements are equal.

    """
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    shared_keys = d1_keys.intersection(d2_keys)
    unique_to_d1 = {k: d1[k] for k in d1_keys - d2_keys}
    unique_to_d2 = {k: d2[k] for k in d2_keys - d1_keys}
    conflicting = {k: (d1[k], d2[k]) for k in shared_keys if d1[k] != d2[k]}
    identical = {k: d1[k] for k in shared_keys if d1[k] == d2[k]}
    return {
        "unique-to-d1": unique_to_d1,
        "unique-to-d2": unique_to_d2,
        "conflicting": conflicting,
        "identical": identical,
    }


def get_smaller_dicts(d: defaultdict | dict, *, max_keys: int) -> list[dict]:
    """From the given dictionary construct smaller dictionaries each with a number of keys equal
    to max_keys or less, and return a list of these smaller dictionaries.
    """
    num_keys = len(d.keys())

    if num_keys <= max_keys:
        logger.warning(
            f"Received <<num_keys>> = {num_keys}, <<max_keys>> = {max_keys}. Nothing "
            f"to do. Returning an empty list."
        )
        return []

    all_keys = list(d.keys())
    key_sets = listutils.get_smaller_lists(all_keys, max_items=max_keys)
    result = []

    for key_set in key_sets:
        smaller = {k: d[k] for k in key_set}
        result.append(smaller)

    return result


def get_pure_dict(d: defaultdict | dict) -> dict:  # Recursive.
    """Receives a defaultdict or dict with possibly nested defaultdicts and returns a pure dict
    object having the same k-v pairs."""
    new = {}
    for k, v in d.items():
        if isinstance(v, dict) or isinstance(v, defaultdict):
            new[k] = get_pure_dict(v)
        else:
            new[k] = v
    return new


def get_defaultdict():
    return defaultdict(get_defaultdict)


def get_key_for_max_value(d: dict) -> Any:
    """Return the key whose value is the maximum."""
    return max(d, key=d.get)


def get_keys_for_largest_n_values(d: dict, *, n: int = 1) -> list:
    """Return a list of the <n> keys whose mapped values are largest in the given dictionary.

    NOTE: The order of the returned keys is not guaranteed.
    """
    if n > 1:
        z = heapq.nlargest(n, d.items(), key=lambda i: i[0])
        return [i[0] for i in z]
    else:
        return [get_key_for_max_value(d)]


def get_string(
    d: dict | defaultdict, n: int = 0, num_tabs: int = 0
) -> str:  # Recursive.
    """Return a nicely-formatted string showing each kv pair in the given dictionary. Recursive."""

    # Convert d to type dict, but since get_pure_dict is recursive do this only on the first call.

    if n == 0 and isinstance(d, defaultdict):
        d_ = get_pure_dict(d)
    else:
        d_ = {k: v for k, v in d.items()}

    s = ""
    tabs = (n + num_tabs + 1) * "\t"
    for k, v in d_.items():
        if isinstance(v, dict):
            s += f"\n{tabs}{k}: {get_string(v, n + 1)}"  # Recursive call.
        else:
            s += f"\n{tabs}{k}: {v}"
    return s


def invert(d: dict) -> None:
    """Invert the kv pairs of the given dictionary, in place. If there are duplicated values
    only the last kv pair will be inverted to a vk pair and the other kv pairs will vanish.
    """
    orig_keys = list(d.keys())
    d_ = {v: k for k, v in d.items()}

    for v in list(d_.keys()):
        d[v] = d_[v]

    for k in orig_keys:
        d.pop(k)


def invert_strict(d: dict) -> None:
    """Similar to invert, which see, but if any values are duplicated raise ValueError."""
    if listutils.has_duplicates(list(d.values())):
        err_msg = "At least two keys map to the same value."
        raise ValueError(err_msg)
    invert(d)


def reorder_by_key(d: dict, *, reverse: bool = False) -> None:
    """Reorder the kv pairs of d by key, in place."""
    d_ = dict(sorted(d.items(), reverse=reverse))  # local copy, sorted

    for k_, v_ in d_.items():
        d.pop(k_)
        d[k_] = v_


def reorder_by_value(d: dict, reverse: bool = False) -> None:
    sorted_items = sorted(d.items(), key=lambda x: x[1], reverse=reverse)
    # Directly build the sorted dictionary
    d.clear()  # Clear the contents of the original dictionary
    d.update(sorted_items)  # Update the original dictionary with sorted items


def replace_key(d: dict, *, k_new: Any, k_old: Any) -> dict:
    """Return a dictionary identical to the given dictionary but with k_old replaced by k_new."""
    try:
        d[k_new] = d.pop(k_old)  # noqa
    except KeyError:
        pass
    return d


def set_all_values(d: dict, *, value: Any) -> dict:
    """Set all values of the given dictionary d to the given parameter val."""
    return {k: value for k in d}



"""Define utility functions for working with lists.
"""
from __future__ import annotations

import logging
from typing import Any

logger = logging.getLogger(__name__)


def diminish(list_a: list, *, wrt: list) -> None:
    """Diminish the first given list with respect the second one, in place. Similar to
    get_common_elements, which see. After this call, the first given list (list_a) will contain
    only those elements in the original list that are also present in the second list (wrt).
    Non-commutative.
    """
    a_ = [i for i in list_a]  # local copy
    for i in a_:
        if i in wrt:
            continue
        else:
            list_a.pop(list_a.index(i))


def filter_by_index(lst: list[list], indexes: list[int]):
    """Receives a list of lists and returns a list of lists except the returned sublist elements
    are selected by index. Parameter indexes provides the selected indexes."""
    return [[j[i] for i in indexes] for j in lst]


def get_all_pairs(given: list) -> list[tuple]:
    """Return a list of all possible unordered pairs of the elements in the given list.

    Example:
        planets = ["sun", "mercury", "venus", "earth"]
        get_all_pairs(planets)

        1...... ('sun', 'mercury')
        2...... ('sun', 'venus')
        3...... ('sun', 'earth')
        4...... ('mercury', 'venus')
        5...... ('mercury', 'earth')
        6...... ('venus', 'earth')
    """
    pairs = []
    for i in range(len(given) - 1):
        sublist = given[i + 1 :]
        for j in sublist:
            pairs.append((given[i], j))
    return pairs


def get_common_elements(list_a: list, list_b: list) -> list:
    """Return a list of elements that are present in both of the given lists. Similar to
    diminish, which changes the first list in place. Commutative."""
    return list(set([i for i in list_a if i in list_b]))


def get_duplicates(list_a: list) -> list:
    """Return a list of the elements that appear at least twice in the given list."""
    if not has_duplicates(list_a):
        return []

    duplicates = set()
    seen = set()

    for i in list_a:
        if i not in seen:
            seen.add(i)
        else:
            duplicates.add(i)

    return list(set(duplicates))


def get_first_not_none_from_end(list_: list) -> Any | None:
    """Return the first element in the given list that is not None starting from the end of the
    list. If there is not such element, return None."""
    list_ = [i for i in list_ if i is not None]
    if len(list_) > 0:
        return list_[-1]


def get_first_not_none_from_start(list_: list) -> Any | None:
    """Return the first element in the given list that is not None starting from the beginning of
    the list."""
    list_ = [i for i in list_ if i is not None]
    if len(list_) > 0:
        return list_[0]


def get_missing_elements(list_a: list, *, wrt: list) -> list:
    """Return a list of the elements in first list that are missing from the second list.
    Non-commutative."""
    return list(set(list_a) - set(wrt))


def get_smaller_lists(given: list, *, max_items: int) -> list[list]:
    """Divide the given list into smaller lists, each with max_items or less.
    """
    num_items = len(given)

    if num_items <= max_items or 0 >= max_items:
        logger.warning(
            f"Received <<num_keys>> = {num_items}, <<max_keys>> = {max_items}. Nothing "
            f"to do. Returning an empty list."
        )
        return []

    return [given[i : i + max_items] for i in range(0, num_items, max_items)]


def get_string(
    list_: list, end_line: bool = False, delimiter: str | None = None
) -> str:
    """Convert the given list to a string."""
    delimiter = ", " if delimiter is None else delimiter
    string = delimiter.join([str(i) if not isinstance(i, str) else i for i in list_])
    if end_line and not string.endswith("\n"):
        string += "\n"
    return string


def has_duplicates(list_: list) -> bool:
    """Return True if any element of the given list (assumed flat) is duplicated."""
    return len(list_) != len(set(list_))


def scalar_multiply(c: int | float, l_: list) -> list:
    """Multiply each element of the given list by <c> and return the new list."""
    return [c * i for i in l_]


def vector_add(l_: list[list]) -> list:
    """Return a list where each element is the element-wise sum of the given lists, which must
    all be of the same length."""
    return [sum(i) for i in zip(*l_)]



"""Define utility methods for working with strings.
"""
from __future__ import annotations

import re


def get_all_numbers(s: str) -> list[str]:
    """Return a list of all substrings (left-to-right) in s that contain only digits."""
    return re.findall(r"\d+", s)


def get_alphanum_split(s: str) -> tuple[str, str]:
    """Return a tuple of two lists, one consisting of the digits in s (as strings) and the other
    the letters in s, maintaining the left-to-right order in each."""
    return get_digits(s), get_letters(s)


def get_digits_as_list(s: str) -> list[str]:
    """Return the characters from string s that are digits, as list of strings."""
    z = [i for i in get_digits(s)]
    return z


def get_digits(s: str) -> str:
    """Return the characters from string s that are digits, as a string."""
    return "".join([c for c in s if c.isdigit()])


def get_letters(s: str) -> str:
    """Return the characters from string s that are letters, as a string."""
    return "".join([c for c in s if c.isalpha()])


def get_ordinal(n: int) -> str:
    """Return the ordinal of the given number as a string."""
    suffixes = {1: "st", 2: "nd", 3: "rd"}
    if 10 <= n % 100 <= 20:
        suffix = "th"
    else:
        suffix = suffixes.get(n % 10, "th")
    return str(n) + suffix


def is_left_substring(str1: str, str2: str) -> bool:
    """Return True if str2 is (1) at least three characters in length and (2) is a substring of
    str1, starting from the left-most character and ignoring case. Not commutative."""
    return len(str2) >= 3 and str1.lower().startswith(str2.lower())


def multi_replace(s: str, *, rmap: dict) -> str:
    """Return a new string in which the replacements specified by the given dictionary (rmap) have
    been applied sequentially starting with source string (s).
    """
    for k, v in rmap.items():
        s = s.replace(k, v)
    return s


def remove_text_inside_brackets(text: str, brackets: str = "()[]") -> str:
    """https://stackoverflow.com/questions/14596884/remove-text-between-and"""
    count = [0] * (len(brackets) // 2)  # count open/close brackets
    saved_chars = list()
    queued_chars = list()  # an intermediate list

    for character in text:
        for i, b in enumerate(brackets):
            if character == b:  # found bracket
                kind, is_closed = divmod(i, 2)
                count[kind] += (-1) ** is_closed  # `+1`: open, `-1`: close

                # If this bracket is never closed, we want to keep it.

                if count[kind] == 1:  # open bracket
                    queued_chars.append(character)

                if count[kind] < 0:  # unbalanced bracket
                    count[kind] = 0  # keep it
                else:  # found bracket to remove
                    queued_chars = []
                    break
        else:  # character is not a [balanced] bracket
            if not any(count):  # outside brackets
                saved_chars.append(character)
            else:
                queued_chars.append(character)

    saved_chars += queued_chars

    return str().join(saved_chars).strip()
